// Géry Casiez

import { io } from "socket.io-client";
let StateMachine = require('javascript-state-machine');

import { OneEuroFilter } from '1eurofilter'

let filteringEnabled = true;
let frequency = 60; // Hz
let mincutoff = 1.0; // Hz
let beta = 0;      
let dcutoff = 1.0; 

// For filtering x and y we need one filter for each axis
let fx = new OneEuroFilter(frequency, mincutoff, beta, dcutoff);
let fy = new OneEuroFilter(frequency, mincutoff, beta, dcutoff);

// To generate the random targets
function getRandomInt(min, max) {
    return Math.floor(Math.random() * (max - min)) + min;
}

const status = document.getElementById("status");
const messages = document.getElementById("messages");

const canvas = document.getElementById('canvas');
const context = canvas.getContext('2d');
canvas.width = window.innerWidth;
canvas.height = window.innerHeight; 

// Speed vertical line
const speedline = document.getElementById('speed');


// One euro filter settings
const fcminslider = document.getElementById('fcmin');
fcminslider.value = mincutoff * 100;
const fcminval = document.getElementById('fcminval');
fcminval.innerHTML = mincutoff + " Hz";

fcminslider.addEventListener('input', () => {
    mincutoff = fcminslider.value/100.0;
    fx.setMinCutoff(mincutoff);
    fy.setMinCutoff(mincutoff);
    fcminval.innerHTML = mincutoff + " Hz";
});

const betaslider = document.getElementById('beta');
betaslider.value = beta * 1000;
const betaval = document.getElementById('betaval');
betaval.innerHTML = beta;

betaslider.addEventListener('input', () => {
    beta = betaslider.value/1000.0;
    fx.setBeta(beta);
    fy.setBeta(beta);
    betaval.innerHTML = beta;
});

// Sigmoid function parameters
const sigmoidsettings = document.getElementById('sigmoidsettings'); 
let Gmin_value = 2.1;
let Gmax_value = 7;
let V1_value = 24.1;
let V2_value = 36.8;

// Position of the cursor
let x = canvas.width/2;
let y = canvas.height/2;
let lastx = x;
let lasty = y;

let xnoisy = x;
let ynoisy = y;

// Targets to display
let targets = [];

for (let i=0; i<20;i+=1) {
    let r = getRandomInt(20, 70); // radius 
    let x = getRandomInt(r, canvas.width-r);
    let y = getRandomInt(r, canvas.height-r);
    let c = "#" + Math.floor(Math.random()*16777215).toString(16); // color
    targets.push({x: x, y: y, r: r, c: c, visible: true});
}

const appendMessage = (content) => {
    const item = document.createElement("li");
    item.textContent = content;
    messages.appendChild(item);
};

const socket = io({
});

socket.on("connect", () => {
    socket.emit("registerDesktopID", socket.id);
    status.innerText = "Connected";
    appendMessage(`event: connect | session id: ${socket.id}`);
});

socket.on("connect_error", (err) => {
    appendMessage(`event: connect_error | reason: ${err.message}`);
});

socket.on("disconnect", (reason) => {
    status.innerText = "Disconnected";
    appendMessage(`event: disconnect | reason: ${reason}`);
});

socket.on("orientation", (angle1, angle2) => {

    
    checkBounds();
    redraw ();
});

socket.on("touchstart", (xin, yin) => {

    redraw ();
});

socket.on("touchmove", (xin, yin) => {

    checkBounds();
    redraw ();
});

socket.on("touchend", () => {

    redraw ();
});

// To keep the pointer in the canvas
function checkBounds() {
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    if (x > canvas.width) x = canvas.width;
    if (y > canvas.height) y = canvas.height;
}

function clearScreen() {
    context.clearRect(0, 0, canvas.width, canvas.height);
}

function isInside(x, y, t) {
    return Math.sqrt((x-t.x)*(x-t.x)+(y-t.y)*(y-t.y)) <= t.r;
  }

function click() {
    console.log("click");
    for (let t of targets) {
        if (isInside(x, y, t)) t.visible = false;
    }
}

// Draw all targets and the pointer
function redraw () {
    clearScreen();
    drawTargets(context);
    drawPointer(context, x, y, 'black');
    drawPointer(context, xnoisy, ynoisy, 'green');
}

function drawPointer(ctx, x, y, color) {
    let size = 30;
    let width = 2;
    
    ctx.strokeStyle = color;
    ctx.lineWidth = width;
    
    ctx.beginPath();
    ctx.moveTo(x, y-size/2);
    ctx.lineTo(x, y+size/2);
    ctx.moveTo(x-size/2, y);
    ctx.lineTo(x+size/2, y);
    ctx.stroke();
}

function drawTargets(ctx) {
    for (let t of targets) {
        if (t.visible) {
            ctx.beginPath();
            ctx.arc(t.x, t.y, t.r/2, 0, 2*Math.PI);
            ctx.fillStyle = t.c;
            ctx.strokeStyle = "#000000";
            ctx.fill();
            ctx.stroke();
        }
    }
}

function sigmoid(speed, Gmin, Gmax, V1, V2) {
    Gmin = Number(Gmin);
    Gmax = Number(Gmax);
    V1 = Number(V1);
    V2 = Number(V2);
    if (speed <= V1) return Gmin;
    if (speed >= V2) return Gmax;
    return Gmin + (Gmax - Gmin) / (V2 - V1) * (speed - V1);
}

// Interaction with the SVG for the transfer function
function addToSVG(svg) {
    
    svg.addEventListener('mousedown', startDrag);
    svg.addEventListener('mousemove', drag);
    svg.addEventListener('mouseup', endDrag);
    svg.addEventListener('mouseleave', endDrag);
    svg.addEventListener('touchstart', startDrag);
    svg.addEventListener('touchmove', drag);
    svg.addEventListener('touchend', endDrag);
    svg.addEventListener('touchleave', endDrag);
    svg.addEventListener('touchcancel', endDrag);

    var S1 = document.getElementById("S1");
    var S2 = document.getElementById("S2");
    var S3 = document.getElementById("S3");

    var Gmin = document.getElementById("Gmin");
    var Gmax = document.getElementById("Gmax");

    function getMousePosition(evt) {
      var CTM = svg.getScreenCTM();
      if (evt.touches) { evt = evt.touches[0]; }
      return {
        x: (evt.clientX - CTM.e) / CTM.a,
        y: (evt.clientY - CTM.f) / CTM.d
      };
    }

    var selectedElement, offset, transform;

    function initialiseDragging(evt) {
        offset = getMousePosition(evt);

        // Make sure the first transform on the element is a translate transform
        var transforms = selectedElement.transform.baseVal;

        if (transforms.length === 0 || transforms.getItem(0).type !== SVGTransform.SVG_TRANSFORM_TRANSLATE) {
          // Create an transform that translates by (0, 0)
          var translate = svg.createSVGTransform();
          translate.setTranslate(0, 0);
          selectedElement.transform.baseVal.insertItemBefore(translate, 0);
        }

        // Get initial translation
        transform = transforms.getItem(0);
        offset.x -= transform.matrix.e;
        offset.y -= transform.matrix.f;
    }

    function startDrag(evt) {
      if (evt.target.classList.contains('draggable')) {
        selectedElement = evt.target;              

        initialiseDragging(evt);
      } else if (evt.target.parentNode.classList.contains('draggable-group')) {
        selectedElement = evt.target.parentNode;
        initialiseDragging(evt);
      }
    }

    function drag(evt) {
      if (selectedElement) {
        evt.preventDefault();
        var coord = getMousePosition(evt);

        if (coord.x < 5.5) coord.x = 5.5;
        if (coord.x > 78) coord.x = 78;
        if (coord.y < 5) coord.y = 5;
        if (coord.y > 39) coord.y = 39;


        if (selectedElement.id == "P1") {
            if (coord.x >  S2.x2.baseVal.value) coord.x = S2.x2.baseVal.value;
            S1.y1.baseVal.value = coord.y;
            S1.x2.baseVal.value = coord.x;
            S1.y2.baseVal.value = coord.y;

            S2.x1.baseVal.value = coord.x;
            S2.y1.baseVal.value = coord.y;

            Gmin.y.baseVal[0].value = coord.y;
            Gmin_value = ((39-coord.y)*10/33).toFixed(1);
            Gmin.textContent = "" + Gmin_value;

            V1_value = ((coord.x - 5.5)).toFixed(1);
        }

        if (selectedElement.id == "P2") {
            if (coord.x <  S1.x2.baseVal.value) coord.x = S1.x2.baseVal.value;
            S2.x2.baseVal.value = coord.x;
            S2.y2.baseVal.value = coord.y;

            S3.x1.baseVal.value = coord.x;
            S3.y1.baseVal.value = coord.y;
            S3.y2.baseVal.value = coord.y;

            Gmax.y.baseVal[0].value = coord.y;
            Gmax_value = ((39-coord.y)*10/33).toFixed(1)
            Gmax.textContent = "" + Gmax_value;

            V2_value = ((coord.x - 5.5)).toFixed(1);
        }

        transform.setTranslate(coord.x - offset.x, coord.y - offset.y);
        updateSigmoidSettings();
      }
    }

    function endDrag(evt) {
      selectedElement = false;
    }
}

function updateSigmoidSettings() {
    sigmoidsettings.innerHTML = "Gmin = " + Gmin_value + " Gmax = " + Gmax_value + " V1 = " + V1_value + " V2 = " + V2_value;
}

// Gaussian noise
// https://stackoverflow.com/questions/25582882/javascript-math-random-normal-distribution-gaussian-bell-curve
function gaussianRandom(mean=0, stdev=1) {
    const u = 1 - Math.random(); // Converting [0,1) to (0,1]
    const v = Math.random();
    const z = Math.sqrt( -2.0 * Math.log( u ) ) * Math.cos( 2.0 * Math.PI * v );
    // Transform to the desired mean and standard deviation:
    return z * stdev + mean;
}

document.addEventListener("DOMContentLoaded", function() {
    const svgeditor = document.getElementById("tfeditor");
    addToSVG(svgeditor);
    updateSigmoidSettings();
    redraw ();
});